# -*- coding: utf-8 -*-
#
# ox4ucs
#  config registry module to update ox configuration
#
# Copyright (C) 2012-2019 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import os
import re
import grp
import pwd
import stat
import syslog
try:
	from typing import List, Tuple
except ImportError:
	pass

DIR_CONFIG = '/opt/open-xchange/etc'
UCR_PREFIX = 'ox/cfg/'
RE_CFGOPTION = '^\s*%s\s*[=:]'
RE_WARNING = re.compile('^# Warning: the value .*? has been set via UCR variable')
MSG_WARNING = '# Warning: the value "%s" has been set via UCR variable "%s" - please alter the UCR variable instead\n'
RE_UCRVAR = re.compile('@%@(.*?)@%@')
RE_FILENAME = re.compile('@&@(.*?)@&@')
FILES_ONLY_READABLE_BY_OX = [
	'/opt/open-xchange/etc/authplugin.properties',
]

KEY = 0
VALUE = 1
REPLACED = 2

syslog.openlog(ident='UCR', facility=syslog.LOG_SYSLOG)


def get_line(option, ucrkey, configRegistry, lineformat):
	"""
	this function loads the UCR value for the specified option
	and replaces UCR placeholders with its current value.
	"""
	value = lineformat % (option, configRegistry.get(ucrkey))
	matches = RE_UCRVAR.findall(value)
	for key in matches:
		value = value.replace('@%%@%s@%%@' % key, configRegistry.get(key, ''))
	matches = RE_FILENAME.findall(value)
	for key in matches:
		try:
			content = open(key, 'r').readline().rstrip('\n\r')
		except IOError:
			content = ''
		value = value.replace('@&@%s@&@' % key, content)
	return value


def handler(configRegistry, changes):
	"""
	main handler
	"""
	handle_files_etc_ox(configRegistry, changes)


def handle_files_etc_ox(configRegistry, changes):
	"""
	handle UCR variable changes with prefix UCR_PREFIX and alter ox
	config files in DIR_CONFIG if neccessary.
	UCR config option style:    ox/cfg/groupware/attachment.properties/MAX_UPLOAD_SIZE=10485760
	                          <PREFIX><------------FILENAME----------><----Option---->=<-Value->
	The UCR variable value may contain UCR placeholders (@%@variable/name@%@).
	The UCR variable value may contain placeholders for files (@&@/etc/fstab@&@).
	"""

	def iter_split_key(key):  # type: (str) -> Tuple[str, List[str]]
		"""
		Split UCR key after each slash to support OX config options with slashes
		and config files in subdirectories (==> filename with slash). If the
		first part of the UCR variable matches to an existing filename, the rear
		part is considered as config option.
		"""
		items = key.split('/')
		# handle keys containing '/': search for existing file
		for index in range(3, len(items)):
			filename = os.path.join(DIR_CONFIG, *items[2:index])
			if os.path.isfile(filename):
				# join rest of key items, so the key is not split anymore
				items = items[:2] + items[2:index] + ['/'.join(items[index:])]
				break
		else:
			# this shouldn't happen anymore, but keeping it because of paranoia
			# fall back to old behavior: '/' is not allowed in UCR keys -> the filename is between ox/cfg/ and the last '/'
			filename = os.path.join(DIR_CONFIG, '/'.join(items[2:-1]))
		return filename, items

	changed_options = {}
	# check all ox UCR variables
	for key in [x for x in configRegistry.keys() if x.startswith(UCR_PREFIX)]:
		# find UCR placeholder in variable values (==> e.g. @%@ldap/server/name@%@)
		matches = RE_UCRVAR.findall(configRegistry.get(key))
		# if key is in list of changed UCR variables or
		# the UCR variable value contains a UCR placeholder that is listed in changed UCR variables list then...
		if key in changes or filter(lambda x: x in changes, matches):
			# ...prepare UCR variable and remember filename, key and value
			filename, items = iter_split_key(key)
			if filename not in changed_options:
				changed_options[filename] = {}
			changed_options[filename][items[-1]] = [key, configRegistry.get(key), False]  # key, value, replaced

	# check for removed UCR variables
	for key in [x for x in changes.keys() if x.startswith(UCR_PREFIX) and x not in configRegistry]:
		filename, items = iter_split_key(key)
		if filename not in changed_options:
			changed_options[filename] = {}
		changed_options[filename][items[-1]] = [key, None, False]  # key, value, replaced

	if changed_options:
		handle_files(configRegistry, changed_options, '%s=%s\n')

		# chmod files with passwords
		for ox_readable_file in [fn for fn in FILES_ONLY_READABLE_BY_OX if fn in changed_options]:
			make_file_readable_only_by_ox(ox_readable_file)


def handle_files(configRegistry, changed_options, lineformat):
	"""
    Update specified files based on data structure "changed_options":

	changed_options = { '/opt/open-xchange/etc/groupware/attachment.properties': { 'mykey':  [ 'ox/cfg/groupware/attachment.properties/mykey', 'the_value', False ],
																				   'thekey': [ 'ox/cfg/groupware/attachment.properties/thekey', 'some_value', False ],
																				 },
																				 ...
					  }
	"""
	# iterate over all filenames specified in UCR variables
	for fn in changed_options.keys():
		# ignore UCR variable if file does not exist
		if not os.path.exists(fn):
			continue

		# read content of file; on error, print msg and ignore file
		try:
			lines = open(fn, 'r').readlines()
		except IOError as exc:
			syslog.syslog('module ox-config.py: cannot read {!r}: {}'.format(fn, exc))
			continue

		file_changed = False
		# iterate over all files and replace line if matching UCR variable does exist
		i = 0
		while i < len(lines):
			for cfgitem in changed_options[fn]:
				if re.match(RE_CFGOPTION % cfgitem, lines[i]):
					# matching line found...

					# ...mark option as found...
					changed_options[fn][cfgitem][REPLACED] = True

					if changed_options[fn][cfgitem][VALUE] is not None:
						# ...and UCR variable does still exist, so current line will be replaced
						lines[i] = get_line(cfgitem, changed_options[fn][cfgitem][KEY], configRegistry, lineformat)
						# check if warning text is missing...
						if (i == 0) or (i > 0 and not RE_WARNING.match(lines[i-1])):
							lines.insert(i, MSG_WARNING % (cfgitem, changed_options[fn][cfgitem][KEY]))
							i += 1
						file_changed = True
					else:
						# ...and UCR variable has been removed, so warning will be removed
						if (i > 0) and RE_WARNING.match(lines[i-1]):
							del lines[i-1]
							i -= 1
							file_changed = True
			i += 1

		# check for unmatched options
		for cfgitem in changed_options[fn]:
			if not(changed_options[fn][cfgitem][REPLACED]):
				# add them at the bottom of the file as new entry
				file_changed = True
				lines.extend([
					'\n',
					MSG_WARNING % (cfgitem, changed_options[fn][cfgitem][KEY]),
					get_line(cfgitem, changed_options[fn][cfgitem][KEY], configRegistry, lineformat),
				])

		# update file only if changes have been made
		if file_changed:
			new_fn = '%s.NEW' % fn
			try:
				# write to temporary file and perform an atomic rename
				old_mode = os.stat(fn).st_mode
				old_uid = os.stat(fn).st_uid
				old_gid = os.stat(fn).st_gid
				fd = open(new_fn, 'w', old_mode)
				# force old permissions
				os.chmod(new_fn, old_mode)
				os.chown(new_fn, old_uid, old_gid)
				fd.write(''.join(lines))
				fd.close()
				os.rename(new_fn, fn)
			except IOError as exc:
				syslog.syslog('module ox-config.py: cannot write {!r}: {}'.format(fn, exc))


def make_file_readable_only_by_ox(filename):
	"""
	chmod 640 filename; chown open-xchange:open-xchange filename

	:param filename: str
	:return: None
	"""
	try:
		ox_uid = pwd.getpwnam('open-xchange').pw_uid
		ox_gid = grp.getgrnam('open-xchange').gr_gid
		os.chmod(filename, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP)
		os.chown(filename, ox_uid, ox_gid)
	except IOError as exc:
		syslog.syslog('module ox-config.py: cannot "chmod 640" or "chown open-xchange:open-xchange" the file {!r}: {}'.format(filename, exc))
