# -*- coding: utf-8 -*-

import pickle
import os
import sys
import time
import random
import string
from traceback import format_exc
import listener
from functools import wraps
from errno import ENOENT, EAGAIN, EACCES
from fcntl import lockf, LOCK_EX, LOCK_NB
from io import StringIO
from subprocess import Popen, PIPE, STDOUT
from csv import DictReader, reader as csv_reader
from pwd import getpwnam
import sqlite3
import stat
import locale
import inspect
import tzlocal
from univention.ox.oxlogger import OXLogger


default_context = int(listener.configRegistry.get('ox/context/id', 10))
__credentials_cache = {}
__master_password = ''


SPOOL_DIR_PREFIX = '/var/spool/univention-directory-listener'
CACHE_DIR_PREFIX = '/var/cache/univention-ox'
OX_BIN_DIRECTORY = '/opt/open-xchange/sbin'

OX_MASTER_ADMIN = "oxadminmaster"
OX_MASTER_SECRET = "/etc/ox-secrets/master.secret"

CSV_DELIMITER = ','
CSV_QUOTECHAR = '"'

CSV_USER_ROW_USERNAME = 'Name'
CSV_USER_ROW_UID = 'Id'

CSV_GROUP_ROW_GROUPNAME = 'name'
CSV_GROUP_ROW_ID = 'id'
CSV_GROUP_ROW_DISPLAYNAME = 'displayname'
CSV_GROUP_ROW_MEMBERS = 'members'

TIMEZONE_SUBSTITUTES_PATH = '/usr/share/univention-ox/timezones.csv'


class ListenerTools_ExceptionBase(Exception):
	pass


class ListenerTools_Unconfigured(ListenerTools_ExceptionBase):
	pass


class ListenerTools_RuntimeError(ListenerTools_ExceptionBase):
	def __init__(self, msg, exit_code=None, *args, **kwargs):   # type: (str, Optional[int], Any, Any) -> None
		self.exit_code = exit_code
		self.stderr = kwargs.pop('stderr', '')
		super(ListenerTools_RuntimeError, self).__init__(msg)


def decode_utf8(*str_args):  # type: (*str) -> Callable
	def inner_wrapper(func):  # type: (Callable) -> Callable
		argspec = inspect.getargspec(func)
		argnames = argspec.args
		named_args = (
			'context_id', 'displayname', 'group_id', 'groupname', 'members', 'new_groupname', 'new_username',
			'ox_user_id', 'username'
		)

		def wrapper_func(*args, **kwargs):  # type: (*unicode, **unicode) -> Callable
			if not str_args or not (isinstance(str_args, list), isinstance(str_args, tuple)) or not all(s in named_args for s in str_args):
				raise ValueError('All arguments to decode_utf8 must be in {!r}. Found arguments: {!r}.'.format(named_args, str_args))
			for named_arg in named_args:
				if named_arg in kwargs:
					ori = kwargs[named_arg]
					if isinstance(ori, str):
						kwargs[named_arg] = ori.decode('utf-8')
			new_args = list(args)  # type: List[unicode]
			for num, argname in enumerate(argnames):
				if argname in named_args and argname not in kwargs:
					try:
						if isinstance(args[num], str):
							new_args[num] = args[num].decode('utf-8')
					except IndexError:
						# default=None arg not set
						pass
					args = tuple(new_args)
			return func(*args, **kwargs)
		return wraps(func)(wrapper_func)
	return inner_wrapper


class OXUserIdCache(object):
	IDX_DB_USERNAME = 0
	IDX_DB_CONTEXT = 1
	IDX_DB_USER_ID = 2

	def __init__(self, module_name=None):  # type: (Optional[str]) -> None
		self.logger = OXLogger(module_name)
		self.logger.qualifier = 'OXUserIdCache'
		self.db = None  # type: Optional[sqlite3.Connection]
		self.cursor = None   # type: Optional[sqlite3.Cursor]
		self.setup_database()

	def setup_database(self):  # type: () -> None
		# close open db handle
		if self.db:
			try:
				self.db.close()
			except sqlite3.Error:
				pass
			self.cursor = None
			self.db = None

		# create cache directory if missing
		if not os.path.exists(CACHE_DIR_PREFIX):
			self.logger.info('cache directory %r does not exist' % (CACHE_DIR_PREFIX,))
			with AsRoot():
				os.makedirs(CACHE_DIR_PREFIX)
			self.logger.info('cache directory successfully created')
		if os.path.exists(CACHE_DIR_PREFIX):
			with AsRoot():
				os.chown(CACHE_DIR_PREFIX, getpwnam('listener').pw_uid, 0)
				os.chmod(CACHE_DIR_PREFIX, stat.S_IRWXU | stat.S_IRWXG)

		# open connection
		filename = os.path.join(CACHE_DIR_PREFIX, 'ox-user-id-cache.db')
		if not os.path.exists(filename):
			self.logger.warn('database does not exist - creating new one (filename=%r)' % (filename,))
		with AsRoot():
			self.db = sqlite3.connect(filename)
			os.chown(filename, getpwnam('listener').pw_uid, 0)
			os.chmod(filename, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP)
		self.logger.debug('opened %r successfully' % (filename,))

		self.cursor = self.db.cursor()

		# create table if missing
		# Note: username has been specified first, for performance reasons. "username" values have
		#       more entropy than "context" values, so the index lookup performance will be better.
		self.cursor.execute('CREATE TABLE IF NOT EXISTS user_id_cache (username TEXT, context_id TEXT, ox_user_id TEXT, PRIMARY KEY (username, context_id))')

		# create index if missing
		# Note: due to PRIMARY KEY on "username" and "context", there is already a lookup index for
		#       "username" and "username+context" but *NOT* for "context". This is fixed by creating
		#       a special lookup index for "context".
		self.cursor.execute('CREATE INDEX IF NOT EXISTS idx_context ON user_id_cache (context_id)')

		# save all changes to database
		self.db.commit()

	def __clear_database(self):  # type: () -> None
		assert self.cursor is not None
		assert self.db is not None
		# drop all rows of table
		self.cursor.execute('DELETE FROM user_id_cache')
		self.db.commit()

	def rebuild_cache(self, raise_exceptions=True):  # type: (Optional[bool]) -> None
		'''
		Purge and rebuild cache. If an exception is raised during rebuild by
		e.g. ox_list_context() the database is dropped again. The handler()
		methods have to check via self.rebuild_cache_if_required() if a rebuild
		is neccessary (because the database is empty).
		'''
		assert self.db is not None
		self.logger.warn('rebuilding cache...')
		self.__clear_database()
		self.logger.info('dropped old entries')

		try:
			masterpass = get_credentials(OX_MASTER_SECRET)
			for context in ox_list_context(OX_MASTER_ADMIN, masterpass, logger=self.logger):
				context_id = context['id']
				self.logger.info('processing context %s...' % (context_id,))
				admuser, admpass = get_credentials_for_context(context_id)
				for row in ox_list_user(admuser, admpass, context_id, logger=self.logger):
					username = row[CSV_USER_ROW_USERNAME]
					ox_user_id = row[CSV_USER_ROW_UID]
					self.add(context_id, username, ox_user_id, db_commit=False)
				self.db.commit()
			self.db.commit()
			self.logger.warn('cache rebuild finished')
		except ListenerTools_RuntimeError as ex:
			# database may be incomplete --> dropping content
			self.logger.warn('cache rebuild failed - dropping all entries')
			self.__clear_database()
			if raise_exceptions:
				raise
			else:
				self.logger.warn('cache rebuild: runtime error: {}'.format(ex))

	def rebuild_cache_if_required(self, raise_exceptions=True):  # type: (Optional[bool]) -> None
		'''
		Rebuild the database if the database is empty, e.g. right after first startup or
		if the last rebuild has failed and all entries have been intentionally been dropped.
		'''
		assert self.cursor is not None
		self.cursor.execute('SELECT COUNT(*) FROM user_id_cache')
		entry_count = self.cursor.fetchone()[0]

		if entry_count == 0:
			self.logger.warn('database is empty - starting cache rebuild')
			self.rebuild_cache(raise_exceptions)
		else:
			self.logger.debug('database contains %d entries' % (entry_count,))

	@decode_utf8('context_id', 'username', 'ox_user_id')
	def add(self, context_id, username, ox_user_id, db_commit=True):
		# type: (unicode, unicode, unicode, Optional[bool]) -> None
		assert self.cursor is not None
		assert self.db is not None
		self.cursor.execute('INSERT INTO user_id_cache VALUES(?, ?, ?)', (username, context_id, ox_user_id))
		if db_commit:
			self.db.commit()
		self.logger.debug('added entry: username=%r  context_id=%r  ox_user_id=%r' % (username, context_id, ox_user_id))

	@decode_utf8('context_id', 'username')
	def remove(self, context_id, username):  # type: (unicode, unicode) -> None
		assert self.cursor is not None
		assert self.db is not None
		self.cursor.execute('DELETE FROM user_id_cache WHERE username=? AND context_id=?', (username, context_id))
		self.db.commit()
		self.logger.debug('removed entry: username=%r  context_id=%r)' % (username, context_id))

	@decode_utf8('context_id', 'new_username', 'ox_user_id')
	def rename_user_by_id(self, context_id, new_username, ox_user_id):  # type: (unicode, unicode, unicode) -> None
		'''
		Changes username in database if context and OX user id are specified.
		'''
		assert self.cursor is not None
		assert self.db is not None
		query = 'SELECT username FROM user_id_cache WHERE context_id=? and ox_user_id=?'
		self.cursor.execute(query, (context_id, ox_user_id))
		row = self.cursor.fetchone()
		if not row:
			self.logger.error('cannot rename OX user with ox_user_id=%r and context_id=%r into username=%s - user not found' % (ox_user_id, context_id, new_username))
			return
		old_username = row[0]
		self.cursor.execute('UPDATE user_id_cache SET username=? WHERE ox_user_id=? AND context_id=?', (new_username, ox_user_id, context_id))
		self.db.commit()
		self.logger.debug('renamed OX user %r into %r (context_id=%r  ox_user_id=%r)' % (old_username, new_username, context_id, ox_user_id))

	@decode_utf8('context_id', 'username')
	def query(self, context_id=None, username=None):
		# type: (Union[unicode, None], Union[unicode, None]) -> Dict[str, Dict[str, str]]
		"""
		Returns {context_id: {username: ox_user_id}} for specified context_id and/or username.
		Both argument may be unspecified to return all entries in cache.
		"""
		assert self.cursor is not None
		result = {}  # type: dict
		filter_str = []
		filter_items = []
		if context_id:
			filter_str.append('context_id = ?')
			filter_items.append(context_id)
		if username:
			filter_str.append('username = ?')
			filter_items.append(username)
		query = 'SELECT * from user_id_cache'
		if filter_items:
			query += ' WHERE %s' % (' AND '.join(filter_str),)
		self.logger.debug('starting sqlite query: %r %r' % (query, filter_items))
		self.cursor.execute(query, filter_items)
		rows = self.cursor.fetchall()
		for row in rows:
			result.setdefault(row[self.IDX_DB_CONTEXT], dict())[row[self.IDX_DB_USERNAME]] = row[self.IDX_DB_USER_ID]
		return result

	@decode_utf8('context_id', 'username')
	def query_ox_id(self, context_id, username):  # type: (unicode, unicode) -> str
		'''
		Returns OX user id for specified user in specified OX context.
		'''
		return self.query(context_id=context_id, username=username).get(context_id, {}).get(username)


class ReadonlyOXGroup(object):
	def __init__(self, groupname, group_id, context_id, displayname, members):
		self._groupname = groupname
		self._group_id = group_id
		self._context_id = context_id
		self._members = members
		self._displayname = displayname

	@property
	def groupname(self):
		return self._groupname

	@property
	def group_id(self):
		return self._group_id

	@property
	def context_id(self):
		return self._context_id

	@property
	def displayname(self):
		return self._displayname

	@property
	def members(self):
		return self._members

	@property
	def member_list(self):
		if self._members:
			return self._members.split(',')
		else:
			return []

	def __str__(self):
		return 'OXGroup(%r, oxgid=%s, context=%s, displayname=%r members=%r)' % (
			self._groupname,
			self._group_id,
			self._context_id,
			self._displayname,
			self._members)

	def __repr__(self):
		return str(self)

class OXGroupCache(object):
	IDX_DB_GROUPNAME = 0
	IDX_DB_GROUP_ID = 1
	IDX_DB_CONTEXT = 2
	IDX_DB_DISPLAYNAME = 3
	IDX_DB_MEMBERS = 4

	def __init__(self, module_name=None):
		self.logger = OXLogger(module_name)
		self.logger.qualifier = 'OXGroupCache'
		self.db = None
		self.cursor = None
		self.setup_database()

	def setup_database(self):
		# close open db handle
		if self.db:
			try:
				self.db.close()
			except sqlite3.Error:
				pass
			self.cursor = None
			self.db = None

		# create cache directory if missing
		if not os.path.exists(CACHE_DIR_PREFIX):
			self.logger.info('cache directory %r does not exist' % (CACHE_DIR_PREFIX,))
			with AsRoot():
				os.makedirs(CACHE_DIR_PREFIX)
			self.logger.info('cache directory successfully created')
		if os.path.exists(CACHE_DIR_PREFIX):
			with AsRoot():
				os.chown(CACHE_DIR_PREFIX, getpwnam('listener').pw_uid, 0)
				os.chmod(CACHE_DIR_PREFIX, stat.S_IRWXU | stat.S_IRWXG)

		# open connection
		filename = os.path.join(CACHE_DIR_PREFIX, 'ox-group-cache.db')
		if not os.path.exists(filename):
			self.logger.warn('database does not exist - creating new one (filename=%r)' % (filename,))
		with AsRoot():
			self.db = sqlite3.connect(filename)
			os.chown(filename, getpwnam('listener').pw_uid, 0)
			os.chmod(filename, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP)
		self.logger.debug('opened %r successfully' % (filename,))

		self.cursor = self.db.cursor()

		# create table if missing
		# Note: groupname has been specified first, for performance reasons. "groupname" values have
		#       more entropy than "context" values, so the index lookup performance will be better.
		self.cursor.execute('CREATE TABLE IF NOT EXISTS group_cache (groupname TEXT, group_id TEXT, context_id TEXT, displayname TEXT, members TEXT, PRIMARY KEY (groupname, context_id))')

		# create index if missing
		# Note: due to PRIMARY KEY on "groupname" and "context", there is already a lookup index for
		#       "groupname" and "groupname+context" but *NOT* for "context". This is fixed by creating
		#       a special lookup index for "context".
		self.cursor.execute('CREATE INDEX IF NOT EXISTS idx_context ON group_cache (context_id)')

		# save all changes to database
		self.db.commit()

	def __clear_database(self):
		# drop all rows of table
		self.cursor.execute('DELETE FROM group_cache')
		self.db.commit()

	def rebuild_cache(self, raise_exceptions=True):
		'''
		Purge and rebuild cache. If an exception is raised during rebuild by
		e.g. ox_list_context() the database is dropped again. The handler()
		methods have to check via self.rebuild_cache_if_required() if a rebuild
		is neccessary (because the database is empty).
		'''
		self.logger.warn('rebuilding cache...')
		self.__clear_database()
		self.logger.info('dropped old entries')

		try:
			masterpass = get_credentials(OX_MASTER_SECRET)
			for context in ox_list_context(OX_MASTER_ADMIN, masterpass, logger=self.logger):
				context_id = context['id']
				self.logger.info('processing context %s...' % (context_id,))
				admuser, admpass = get_credentials_for_context(context_id)
				for row in ox_list_group(admuser, admpass, context_id, logger=self.logger):
					groupname = row[CSV_GROUP_ROW_GROUPNAME]
					group_id = row[CSV_GROUP_ROW_ID]
					displayname = row[CSV_GROUP_ROW_DISPLAYNAME]
					members = row[CSV_GROUP_ROW_MEMBERS]
					self.add(context_id, groupname, group_id, displayname, members, db_commit=False)
				self.db.commit()
			self.db.commit()
			self.logger.warn('cache rebuild finished')
		except ListenerTools_RuntimeError as ex:
			# database may be incomplete --> dropping content
			self.logger.warn('cache rebuild failed - dropping all entries')
			self.__clear_database()
			if raise_exceptions:
				raise
			else:
				self.logger.warn('cache rebuild: runtime error: {}'.format(ex))

	def rebuild_cache_if_required(self, raise_exceptions=True):
		'''
		Rebuild the database if the database is empty, e.g. right after first startup or
		if the last rebuild has failed and all entries have been intentionally been dropped.
		'''
		self.cursor.execute('SELECT COUNT(*) FROM group_cache')
		entry_count = self.cursor.fetchone()[0]

		if entry_count == 0:
			self.logger.warn('database is empty - starting cache rebuild')
			self.rebuild_cache(raise_exceptions)
		else:
			self.logger.debug('database contains %d entries' % (entry_count,))

	@decode_utf8('context_id', 'groupname', 'group_id', 'displayname', 'members')
	def add(self, context_id, groupname, group_id, displayname, members, db_commit=True):
		# type: (unicode, unicode, unicode, unicode, unicode, Optional[bool]) -> None
		'''
		add('10', 'MyOXGroupName', '5', 'My new OX group', '3,4,6,34,2342', db_commit=False)
		'''
		self.cursor.execute('INSERT INTO group_cache VALUES(?, ?, ?, ?, ?)', (groupname, group_id, context_id, displayname, members))
		if db_commit:
			self.db.commit()
		self.logger.debug('added entry: groupname=%r  group_id=%r  context_id=%r  displayname=%r  members=%r' % (groupname, group_id, context_id, displayname, members))

	@decode_utf8('context_id', 'groupname')
	def add_from_ox(self, context_id, groupname, admuser, admpass):  # type: (unicode, unicode, str, str) -> None
		'''
		Add group for a specific group and context with data from OX.
		'''
		for row in ox_list_group(admuser, admpass, context_id, searchpattern=groupname, logger=self.logger):
			if row[CSV_GROUP_ROW_GROUPNAME].lower() == groupname.lower():
				self.logger.debug('add_from_ox: groupname=%r  group_id=%r  context_id=%r  displayname=%r  members=%r' % (
					groupname,
					row[CSV_GROUP_ROW_ID],
					context_id,
					row[CSV_GROUP_ROW_DISPLAYNAME],
					row[CSV_GROUP_ROW_MEMBERS]))
				self.add(context_id, groupname, row[CSV_GROUP_ROW_ID], row[CSV_GROUP_ROW_DISPLAYNAME], row[CSV_GROUP_ROW_MEMBERS])

	@decode_utf8('context_id', 'groupname')
	def remove(self, context_id, groupname):  # type: (unicode, unicode) -> None
		self.cursor.execute('DELETE FROM group_cache WHERE groupname=? AND context_id=?', (groupname, context_id))
		self.db.commit()
		self.logger.debug('removed entry: groupname=%r  context_id=%r)' % (groupname, context_id))

	@decode_utf8('context_id', 'groupname')
	def add_members(self, context_id, groupname, members):  # type: (unicode, unicode, Union[str, List[str]]) -> None
		'''
		add_members('10', 'MyOXGroupName', '7,9,12')
		add_members('10', 'MyOXGroupName', ['7', '9', '12'])
		'''
		query = 'SELECT members FROM group_cache WHERE context_id=? and groupname=?'
		self.cursor.execute(query, (context_id, groupname))
		row = self.cursor.fetchone()
		if not row:
			old_members = set()  # type: Set[unicode]
		else:
			old_members = set(row[0].replace(',', ' ').split())
		if isinstance(members, list) or isinstance(members, set):
			old_members.update(members)
		else:
			old_members.update(members.replace(',', ' ').split())
		self.cursor.execute('UPDATE group_cache SET members=? WHERE groupname=? AND context_id=?', (u','.join(old_members), groupname, context_id))
		self.db.commit()

	@decode_utf8('context_id', 'groupname')
	def remove_members(self, context_id, groupname, members):  # type: (unicode, unicode, Union[str, List[str]]) -> None
		'''
		remove_members('10', 'MyOXGroupName', '7,9,12')
		remove_members('10', 'MyOXGroupName', ['7', '9', '12'])
		'''
		query = 'SELECT members FROM group_cache WHERE context_id=? and groupname=?'
		self.cursor.execute(query, (context_id, groupname))
		row = self.cursor.fetchone()
		if not row:
			old_members = set()  # type: Set[unicode]
		else:
			old_members = set(row[0].replace(',', ' ').split())
		if isinstance(members, list) or isinstance(members, set):
			new_members = old_members - set(members)
		else:
			new_members = old_members - set(members.replace(',', ' ').split())
		self.cursor.execute('UPDATE group_cache SET members=? WHERE groupname=? AND context_id=?', (u','.join(new_members), groupname, context_id))
		self.db.commit()

	@decode_utf8('context_id', 'new_groupname', 'group_id')
	def rename_group_by_id(self, context_id, new_groupname, group_id):  # type: (unicode, unicode, unicode) -> None
		'''
		Changes groupname in database if context and OX group id are specified.
		'''
		query = 'SELECT groupname FROM group_cache WHERE context_id=? and group_id=?'
		self.cursor.execute(query, (context_id, group_id))
		row = self.cursor.fetchone()
		if not row:
			self.logger.error('cannot rename OX group with group_id=%r and context_id=%r into groupname=%s - group not found' % (group_id, context_id, new_groupname))
			return
		old_groupname = row[0]
		self.cursor.execute('UPDATE group_cache SET groupname=? WHERE group_id=? AND context_id=?', (new_groupname, group_id, context_id))
		self.db.commit()
		self.logger.debug('renamed OX group %r into %r (context_id=%r  group_id=%r)' % (old_groupname, new_groupname, context_id, group_id))

	@decode_utf8('context_id', 'groupname', 'group_id')
	def query(self, context_id=None, groupname=None, group_id=None):
		# type: (Union[unicode, None], Union[unicode, None], Union[unicode, None]) -> Dict[str, List[ReadonlyOXGroup]]
		"""
		Returns {context_id: [ReadonlyOXGroup]} for specified context_id, groupname and/or group_id.
		All arguments may be unspecified to return all entries in cache.
		"""
		result = {}  # type: dict
		filter_str = []
		filter_items = []
		if context_id:
			filter_str.append('context_id = ?')
			filter_items.append(context_id)
		if groupname:
			filter_str.append('groupname = ?')
			filter_items.append(groupname)
		if group_id:
			filter_str.append('group_id = ?')
			filter_items.append(group_id)
		query = 'SELECT * from group_cache'
		if filter_items:
			query += ' WHERE %s' % (' AND '.join(filter_str),)
		self.logger.debug('starting sqlite query: %r %r' % (query, filter_items))
		self.cursor.execute(query, filter_items)
		rows = self.cursor.fetchall()
		for row in rows:
			result.setdefault(row[self.IDX_DB_CONTEXT], []).append(
				ReadonlyOXGroup(
					row[self.IDX_DB_GROUPNAME],
					row[self.IDX_DB_GROUP_ID],
					row[self.IDX_DB_CONTEXT],
					row[self.IDX_DB_DISPLAYNAME],
					row[self.IDX_DB_MEMBERS]))
		return result

	@decode_utf8('context_id', 'groupname', 'group_id')
	def query_one_context(self, context_id, groupname=None, group_id=None):
		# type: (unicode, Union[unicode, None], Union[unicode, None]) -> List[ReadonlyOXGroup]
		'''
		Returns list of ReadonlyOXGroup in specified OX context for optional arguments (if specified).
		'''
		return self.query(context_id=context_id, groupname=groupname, group_id=group_id).get(context_id, [])


class LDAPObject(dict):
	"""
	LDAPObject behaves like a normal dictionary containing the attribute values of
	a LDAP object. In addition it is possible to read the DN of the LDAP object from
	a member attribute.

	>>> new = LDAPObject(dn='uid=Administrator,cn=users,dc=example,dc=com', uid=['Administrator'])
	>>> new.dn
	'uid=Administrator,cn=users,dc=example,dc=com'
	>>> new['uid']
	['Administrator']
	"""
	def __init__(self, dn, *args, **kwargs):
		self.dn = dn
		dict.__init__(self, *args, **kwargs)


class QueueEntry(object):
	"""
	This class represents one queue entry and will be pickled to disk for offline storage.
	It contains the old and new LDAP object (as "LDAPObject"), the action and the timestamp
	of creation.
	"""
	def __init__(self, dn, new, old, action, timestamp):
		self.new = LDAPObject(dn, new)
		self.old = LDAPObject(dn, old)
		self.action = action
		self.timestamp = timestamp

	@property
	def dn(self):
		"""
		Returns DN of LDAP object (preferably the "new" DN).
		"""
		if self.new and self.new.dn:
			return self.new.dn
		elif self.old and self.old.dn:
			return self.old.dn
		return None


class ListenerQueue(object):
	"""
	Example listener code:

	@ListenerQueue(module_name=name)
	def handler(dn, new, old, action=None):
		if not os.path.exists('/etc/foo.secret'):
			return ListenerTools_RuntimeError('this service is unconfigured')
		if not foo.service_is_available:
			raise ListenerTools_RuntimeError('the service is down')
		foo.do_something(dn, new, old)
	"""
	def __init__(self, module_name=None, stop_on_error=True):
		"""
		Creates a new ListenerQueue and loads existing pickle files into the queue.

		module_name:
			name of the listener module. The spooling directory is derived
			from the module name: '/var/spool/univention-directory-listener/$(module_name)s/'
		stop_on_error:
			If stop_on_error is True, the listener module stops processing queue entries
			on the first error/exception. Otherwise the module will skip the erroneous
			queue entry and continue processing the other one. Errorneous queue entries
			will be reprocessed on the next call of process().
		"""
		global name
		self.stop_on_error = stop_on_error
		if module_name:
			self.module_name = module_name
		else:
			try:
				self.module_name = name
			except NameError:
				raise ListenerTools_Unconfigured('The module_name has not been specified and could not be determined automatically!')
		self.spool_dir = os.path.join(SPOOL_DIR_PREFIX, self.module_name)
		self.queue = []
		self.logger = OXLogger(self.module_name)

		self.reread_queue_from_disk()

	@classmethod
	def get_lock_filename(cls, module_name):
		return os.path.join(SPOOL_DIR_PREFIX, module_name, 'lock.tmp')

	@property
	def queue_size(self):
		return len(self.queue)

	def __call__(self, func):
		def process_pending():
			ent = getpwnam("listener")
			os.seteuid(ent.pw_uid)

			try:
				return self.process(func)
			except ListenerTools_RuntimeError as ex:
				self.logger.warn('Runtime error: {}'.format(ex))
				return

		@wraps(func)
		def wrapper(dn, new, old, action=None):
			self.add(dn, new, old, action)
			return self.process(func)

		wrapper.process_pending = process_pending
		return wrapper

	def add(self, dn, new, old, action=None):
		"""
		Add new entry to queue.
		"""
		# use timestamp of current time as filename
		# this way files/timestamps may be simply sorted later on
		timestamp = '%.10f' % time.time()
		entry = QueueEntry(dn, new, old, action, timestamp)
		self._save_entry(entry)
		self.queue.append(timestamp)

	def process(self, func):
		"""
		Process all available objects in queue. The specified function "func" is
		called with four arguments:
		dn: DN of the LDAP object
		new: new attributes
		old: old attributes
		action: modification type (only if specified during add)
		"""
		result = None
		dn = None

		with Exclusive():
			queue_unprocessed = True
			while queue_unprocessed:
				queue_unprocessed = False

				for timestamp in self.queue:
					self.logger.info('Loading transaction {}'.format(timestamp))
					try:
						entry = self._load_entry(timestamp)
					except IOError as ex:
						if ex.errno != ENOENT:
							self.logger.error('loading queue entry {} failed (errno {})'.format(timestamp, ex.errno))
							raise
						self.reread_queue_from_disk()
						queue_unprocessed = True
						break

					try:
						dn = entry.dn
						result = func(dn, entry.new, entry.old, entry.action)
						self._delete_entry(timestamp)
						self.logger.info('transaction {} processed: dn: {}'.format(timestamp, dn))
					except ListenerTools_RuntimeError as ex:
						self.logger.process('transaction {} deferred: dn: {} {}'.format(timestamp, dn, ex))
						if self.stop_on_error:
							break
					except Exception as ex:
						self.logger.error('handler failed for transaction {}: dn: {}\n{}\n{}'.format(timestamp, dn, format_exc(), ex))
						raise

			# some entries might have been removed, so the queue is reread
			self.reread_queue_from_disk()

		return result

	def clear_queue(self):
		for fn in os.listdir(self.spool_dir):
			if fn.endswith('.tmp'):
				continue
			os.remove(fn)

	def reread_queue_from_disk(self):
		if not os.path.isdir(self.spool_dir):
			self.queue = []
			return
		self.queue = [fn for fn in os.listdir(self.spool_dir) if not fn.endswith('.tmp')]
		sorted(self.queue)

	def _save_entry(self, entry):
		"""
		Write entry as pickle file to spool directory.
		"""

		# if spool directory is missing then create it
		if not os.path.isdir(self.spool_dir):
			os.makedirs(self.spool_dir, 0700)

		# write pickle file
		fn = os.path.join(self.spool_dir, entry.timestamp)
		fn_tmp = fn + '.tmp'
		with open(fn_tmp, 'wb') as fd:
			os.chmod(fn_tmp, 0600)
			pickle.dump(entry, fd, 2)
			fd.flush()
			os.fsync(fd.fileno())
		os.rename(fn_tmp, fn)

	def _load_entry(self, timestamp):
		"""
		Load pickle object from disk
		"""
		fn = os.path.join(self.spool_dir, timestamp)
		with open(fn, 'rb') as fd:
			return pickle.load(fd)

	def _delete_entry(self, timestamp):
		fn = os.path.join(self.spool_dir, timestamp)
		os.remove(fn)


class AsRoot(object):
	"""
	Temporarily change effective UID to 'root', if it's not already 'root'.
	"""
	def __init__(self):
		self.old_uid = None

	def __enter__(self):  # type: () -> None
		self.old_uid = os.geteuid()
		if self.old_uid != 0:
			listener.setuid(0)

	def __exit__(self, exc_type, exc_value, traceback):  # type: (Type[BaseException], BaseException, types.TracebackType) -> None
		if self.old_uid == os.geteuid():
			return
		listener.unsetuid()


class Exclusive(object):
	def __init__(self, filename=None, module_name=None):
		# type: (Optional[str], Optional[str]) -> None
		self.lock_name = filename
		if module_name:
			self.lock_name = ListenerQueue.get_lock_filename(module_name)
		# use generic lock if no module or filename has been specified
		if self.lock_name is None:
			self.lock_name = os.path.join(SPOOL_DIR_PREFIX, 'tmp', 'ox-lock.tmp')
		self.lock_file = None  # type: Optional[int]

	def __enter__(self):
		# type: () -> None
		try:
			lock_dir = os.path.dirname(self.lock_name)
			if not os.path.exists(lock_dir):
				with AsRoot():
					os.makedirs(lock_dir)
					os.chown(lock_dir, getpwnam('listener').pw_uid, 0)
					os.chmod(lock_dir, stat.S_IRWXU)
			lock_file = os.open(self.lock_name, os.O_RDWR | os.O_CREAT, 0600)
			lockf(lock_file, LOCK_EX | LOCK_NB)
			with AsRoot():
				os.fchown(lock_file, getpwnam('listener').pw_uid, 0)
				os.fchmod(lock_file, stat.S_IRUSR | stat.S_IWUSR)
			self.lock_file = lock_file
		except IOError as ex:
			if ex.errno not in (EAGAIN, EACCES):
				raise
			if hasattr(sys, 'argv'):
				argv = ', command: {!r}'.format(' '.join(sys.argv))
			else:
				argv = ', listener'
			raise ListenerTools_RuntimeError("concurrent process (pid: {}{})".format(os.getpid(), argv))

	def __exit__(self, exc_type, exc_value, traceback):
		if self.lock_file:
			os.close(self.lock_file)
			self.lock_file = None


def get_context_admin_user(context_id):  # type: (Union[int, str]) -> str
	if int(context_id) == 10:
		return 'oxadmin'
	else:
		return 'oxadmin-context{}'.format(context_id)


def get_credentials_for_context(context_id):  # type: (Union[int, str]) -> Tuple[str, str]
	context_id = int(context_id)
	if context_id not in __credentials_cache:
		admuser = get_context_admin_user(context_id)
		admpass = get_credentials('/etc/ox-secrets/context{}.secret'.format(context_id))
		__credentials_cache[context_id] = (admuser, admpass)
	return __credentials_cache[context_id]


def get_master_credentials():  # type: () -> Tuple[str, str]
	global __master_password

	if not __master_password:
		__master_password = get_credentials(OX_MASTER_SECRET)
	return OX_MASTER_ADMIN, __master_password


def get_credentials(filename):  # type: (str) -> str
	"""
	Try to read secret file using effective UID 'root'.
	"""
	with AsRoot():
		try:
			with open(filename, 'r') as sfile:
				secret = sfile.read()
		except IOError as ex:
			raise ListenerTools_RuntimeError(str(ex))
	return secret.rstrip()


def ox_list_context(user, password, logger=None):  # type: (str, str, Optional[OXLogger]) -> DictReader
	return _ox_csvreader('listcontext', user, password, logger=logger)


def ox_list_user(user, password, context, searchpattern=None, ignorecase=False, logger=None):
	# type: (str, str, str, Optional[str], Optional[bool], Optional[OXLogger]) -> DictReader
	return _ox_csvreader('listuser', user, password, context, searchpattern, ignorecase, logger=logger)


def ox_list_group(user, password, context, searchpattern=None, logger=None):
	# type: (str, str, str, Optional[str], Optional[OXLogger]) -> DictReader
	return _ox_csvreader('listgroup', user, password, context, searchpattern, logger=logger)


def ox_list_resource(user, password, context, searchpattern=None, logger=None):
	# type: (str, str, str, Optional[str], Optional[bool], Optional[OXLogger]) -> DictReader
	return _ox_csvreader('listresource', user, password, context, searchpattern, logger=logger)


def _ox_csvreader(list_type, user, password, context=None, searchpattern=None, ignorecase=False, logger=None):
	# type: (str, str, str, Optional[str], Optional[str], Optional[bool], Optional[OXLogger]) -> DictReader
	"""
	Execute listuser/group, read output and return csv.DictReader.
	"""
	if not logger:
		logger = OXLogger('ox_csvreader')
		logger.stdout = True
	command = [
		os.path.join(OX_BIN_DIRECTORY, list_type),
		'--adminuser=%s' % user,
		'--adminpass=%s' % password,
		'--csv',
	]
	if context:
		command += ['--contextid', '%s' % context]
	if searchpattern:
		command += ['--searchpattern', '%s' % searchpattern]
	if ignorecase:
		command += ['--ignorecase']

	logger.info('command {}'.format(' '.join(_debug_command(command))))
	process = Popen(command, stdout=PIPE, stderr=PIPE)
	out, err = process.communicate()
	if process.returncode:
		logger.info('{} {} returned exitcode {}\nSTDOUT:\n{}\nSTDERR:\n{}'.format(list_type, context, process.returncode, out, err))
		raise ListenerTools_RuntimeError(
			'"{} {}" returned  {}'.format(list_type, context, process.returncode),
			process.returncode,
			stderr=err)

	out = out.decode(locale.getpreferredencoding())  # io.StringIO requires a unicode string
	memfile = StringIO(out)  # csv.DictReader wants a file-like-object
	return DictReader(memfile, delimiter=CSV_DELIMITER, quotechar=CSV_QUOTECHAR)


def ox_call(command, expected_error=None, logger=None):  # type: (List[str], Optional[str], Optional[OXLogger]) -> None
	"""
	Execute command and raise RuntimeError on failures.
	"""
	if not logger:
		logger = OXLogger('ox_call')
		logger.stdout = True

	command = [os.path.join(OX_BIN_DIRECTORY, command[0])] + command[1:]
	logger.info('command {}'.format(' '.join(_debug_command(command))))
	process = Popen(command, stdout=PIPE, stderr=STDOUT)
	stdout, _stderr = process.communicate()
	if process.returncode == 0:
		return
	if expected_error and expected_error in stdout:
		logger.process('ignoring expected error "{}": {}'.format(expected_error, stdout.split('\n')[0].rstrip(' :')))
		return
	logger.error('failed with {}: {}'.format(process.returncode, stdout))
	raise ListenerTools_RuntimeError('"{}" returned  {}: {}'.format(_debug_command(command), process.returncode, stdout), process.returncode)


def _debug_command(command):  # type: (List[str]) -> List[str]
	"""
	Dump command excluding passwords.
	"""
	return [
		c for c in command
		if not c.startswith("--adminpass=") and not c.startswith("--password=")
	]


def listener_is_deactivated(name):  # type: (str) -> bool
	if listener.configRegistry.is_false('ox/listener/enabled', False):
		OXLogger(name).process('listener module deactivated by UCR variable "ox/listener/enabled".')
		return True
	return False


def get_local_timezone():  # type: () -> str
	os_tz = tzlocal.get_localzone().zone
	with open(TIMEZONE_SUBSTITUTES_PATH) as fp:
		reader = csv_reader(fp, delimiter='\t')
		for row in reader:
			if row[0] == os_tz:
				if len(row) > 1:
					return row[1]
				else:
					return row[0]
	return os_tz


def get_disk_free(path):  # type: (str) -> int
	res = os.statvfs(path)
	return res.f_frsize * res.f_bavail / 1024 / 1126  # TODO: why 1126? (copied from ox-lib.sh -> setup_vars ())


def get_random_password(length=15):  # type: (Optional[int]) -> str
	specials='@#$%&*-_+=:,.;?/()'
	pw = list()
	if length >= 4:
		pw.append(random.choice(string.lowercase))
		pw.append(random.choice(string.uppercase))
		pw.append(random.choice(string.digits))
		pw.append(random.choice(specials))
		length -= len(pw)
	pw.extend(random.choice(string.ascii_letters + string.digits + specials) for _x in range(length))
	random.shuffle(pw)
	return ''.join(pw)


def get_new_context_attributes(context_id, domain=None):  # type: (Union[int,str], Optional[str]) -> Dict[str, str]
	name = 'context{}'.format(context_id)
	mapping = ['context{}'.format(context_id), str(context_id)]
	if domain:
		mapping.append(domain)
	else:
		# TODO: IMHO this should be: ucr['mail/hosteddomains'].split()[0] (copied from ox-lib.sh -> setup_vars())
		domain = listener.configRegistry['domainname']
	return {
		'id': int(context_id),
		'name': name,
		'mapping': mapping,
		'timezone': get_local_timezone(),
		'username': get_context_admin_user(context_id),
		'displayname': 'OX Admin',
		'givenname': 'OX',
		'surname': 'Admin',
		'email': '{}@{}'.format(get_context_admin_user(context_id), domain),
		'quota': get_disk_free('/var/oxfilestore'),
		'password': get_random_password(),
	}


if __name__ == '__main__':
	import doctest
	doctest.testmod()
