#!/usr/bin/python2.7
# -*- coding: utf-8 -*-

# Copyright (C) 2019 Univention GmbH <http://www.univention.de/>
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

"""
Create or update a YAML file.
"""

import codecs
import logging
import os.path
import datetime
import shutil
import ruamel.yaml
try:
	from typing import Any, Callable, Dict, Optional
except ImportError:
	pass


logger = logging.getLogger(__name__)


def load_yaml(filename):  # type: (str) -> Dict
	with codecs.open(filename, 'r', 'utf-8') as fp:
		return ruamel.yaml.load(fp, ruamel.yaml.RoundTripLoader, preserve_quotes=True)


def dump_yaml(filename, data, backup=True):  # type: (str, Dict[str, Any], Optional[bool]) -> None
	if backup and os.path.exists(filename):
		spl = filename.rsplit('.', 1)
		if spl[-1].upper() in ('YML', 'YAML'):
			fn, ext = spl
			ext = '.{}'.format(ext)
		else:
			fn = filename
			ext = ''
		backup_fn = '{}_{}{}'.format(fn, datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S'), ext)
		logger.info('Creating backup of %r as %r.', filename, backup_fn)
		shutil.copy2(filename, backup_fn)
	with codecs.open(filename, 'w', 'utf-8') as fp:
		ruamel.yaml.dump(data, fp, Dumper=ruamel.yaml.RoundTripDumper, indent=4, width=400, encoding='utf-8', allow_unicode=True)


def create_or_update(filename, set_data_func, backup=True):
	# type: (str, Callable[[Dict[str, Any]], bool], Optional[bool]) -> None
	"""
	Create or update a YAML file.

	`set_data_func` must be a function accepting a dictionary (the content
	of the YAML file) and returning a boolean indicating whether the data was
	changed. If it is false, the YAML file will not be modified and no
	backup file will be created.

	:param str filename: file to update
	:param callable set_data_func: function(dict) -> bool
	:param bool backup: if a backup of the input file should be made
	:return: True if the data was changed, else False
	:rtype: bool
	"""
	if os.path.exists(filename):
		data = load_yaml(filename)
	else:
		logger.info('Creating new file %r.', filename)
		data = {}
	changed = set_data_func(data)
	if changed:
		dump_yaml(filename, data, backup)
