#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
#  OX for UCS
#  listener module: handling contexts
#
# Copyright (C) 2004-2019 Univention GmbH <http://www.univention.de/>
# and
#     iKu Systems & Services GmbH & Co. KG (http://www.iku-systems.de/)
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

__package__ = ''  # workaround for PEP 366
from univention.ox.oxlogger import OXLogger
from univention.ox.oxlistener import OxListener
from univention.ox.listener_tools import ListenerQueue, OXGroupCache, get_master_credentials, ox_list_context, ox_call, listener_is_deactivated

name = 'ox-context'
description = 'Update Open-Xchange context'
filter = '(objectClass=oxContext)'
attributes = []

CSV_CONTEXT_COL_ID = 'id'
CSV_CONTEXT_COL_FILESTORE = 'filestore_id'

logger = None


def modify_context(new):
	OX_MASTER_ADMIN, admpass = get_master_credentials()
	oxquota = new.get('oxQuota', ['-1'])[0]
	if oxquota == '0':
		oxquota = '-1'
	context_id = new['oxContextIDNum'][0]
	logger.qualifier = context_id
	logger.info('setting quota: {}'.format(oxquota))

	# set quota on context
	command = [
		'changecontext',
		'--adminuser=%s' % OX_MASTER_ADMIN,
		'--adminpass=%s' % admpass,
		'--contextid', context_id,
		'--quota', oxquota,
	]
	ox_call(command, logger=logger)

	# get filestore id for context
	filestoreId = False
	for row in ox_list_context(OX_MASTER_ADMIN, admpass):
		if row[CSV_CONTEXT_COL_ID] == new['oxContextIDNum'][0]:
			filestoreId = row[CSV_CONTEXT_COL_FILESTORE]
	# set quota on filestore
	if filestoreId:
		command = [
			'changefilestore',
			'--adminuser=%s' % OX_MASTER_ADMIN,
			'--adminpass=%s' % admpass,
			'--id', filestoreId,
			'--storesize', oxquota
		]
		ox_call(command, logger=logger)

	# find groups without contextid and create them in new context
	udm_groups = OxListener.get_udm_ox_groups('(!(oxContext=*))')
	context_groups = OxListener.get_ox_groups(context_id)
	for udm_group in [udm_group for udm_group in udm_groups if udm_group.info['name'] not in context_groups]:
		logger.process('creating OX group "{}" in new context {}...'.format(udm_group.info['name'], context_id))
		admuser, admpass = OxListener.get_credentials(context_id)

		group_display_name = udm_group.info.get('description', udm_group.info['name'])  # use description as displayname (or group name as fallback)
		group_display_name = group_display_name[:128]  # OX length limit (Bug #35820)

		command = [
			'creategroup',
			'--displayname={}'.format(group_display_name),
			'--adminuser={}'.format(admuser),
			'--adminpass={}'.format(admpass),
			'--contextid={}'.format(context_id),
			'--name={}'.format(udm_group.info['name']),
		]
		ox_call(command, logger=logger)

		_ox_group_cache = OXGroupCache()
		_ox_group_cache.rebuild_cache_if_required()
		_ox_group_cache.add_from_ox(context_id, udm_group.info['name'], admuser, admpass)


def delete_context(old):
	# find groups with this contextid and modify them to remove it
	context_id = old['oxContextIDNum'][0]
	logger.qualifier = context_id
	udm_groups = OxListener.get_udm_ox_groups('(oxContext={})'.format(context_id), modify=True)

	for udm_group in udm_groups:
		logger.process('modifying UDM group "{}" to remove context {} from it...'.format(udm_group.info['name'], context_id))
		udm_group['oxContext'].remove(context_id)
		udm_group.modify()


@ListenerQueue(name, stop_on_error=False)
def handler(dn, new, old, action=None):
	global logger

	if listener_is_deactivated(name):
		return

	logger = OXLogger(name)

	if new:
		modify_context(new)
		return

	if old and not new:
		delete_context(old)
		return


if __name__ == '__main__':
	handler.process_pending()
