#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
#  OX for UCS
#  listener module: handling resources
#
# Copyright (C) 2004-2019 Univention GmbH <http://www.univention.de/>
# and
#     iKu Systems & Services GmbH & Co. KG (http://www.iku-systems.de/)
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

__package__ = ''  # workaround for PEP 366
import os
import listener
from univention.ox.oxlistener import OxListener
from univention.ox.listener_tools import ListenerQueue, ox_call, listener_is_deactivated, default_context


name = 'ox-resource'
description = 'Resource settings for Open-Xchange'
filter = '(objectClass=oxResourceObject)'
attributes = []

os.environ["LANG"] = listener.configRegistry.get("locale/default", "").split(':')[0]


class OXResourceListener(OxListener):
	def __init__(self, name, old, new):
		super(OXResourceListener, self).__init__(name, old, new)
		self.resource_name = self.get_resource_name()
		self.logger.qualifier = self.resource_name
		self.old_context_id = self.old.get('oxContextIDNum', [default_context])[0]
		self.new_context_id = self.new.get('oxContextIDNum', [default_context])[0]

	def get_resource_name(self):
		if self.new:
			return self.new['cn'][0]
		elif self.old:
			return self.old['cn'][0]
		else:
			return ''

	def _default_create_modify_args(self, email, display_name, description):
		args = [
			'--email={}'.format(email),
			'--available=true',
			'--name={}'.format(self.resource_name),
		]
		if display_name:
			args.append('--displayname={}'.format(display_name))
		if description:
			args.append('--description={}'.format(description))
		return args

	def create_resource(self, context_id):
		self.logger.process('Creating Open-Xchange resource "{}" in context {}.'.format(self.resource_name, context_id))
		command = ['createresource']
		command.extend(self._default_create_modify_args(
			self.new['mailPrimaryAddress'][0],
			self.new.get('displayName', [''])[0],
			self.new.get('description', [''])[0]
		))
		command.extend(self._auth_args(context_id))
		ox_call(command, logger=self.logger)

	def modify_resource(self, context_id):
		self.logger.process('Modifying Open-Xchange resource "{}" in context {}.'.format(self.resource_name, context_id))
		command = ['changeresource']
		command.extend(self._default_create_modify_args(
			self.new['mailPrimaryAddress'][0],
			self.new.get('displayName', [''])[0],
			self.new.get('description', [''])[0]
		))
		command.extend(self._auth_args(context_id))
		ox_call(command, logger=self.logger)

	def remove_resource(self, context_id):
		self.logger.process('Deleting Open-Xchange resource "{}" in context {}.'.format(self.resource_name, context_id))
		command = [
			'deleteresource',
			"--name=%s" % self.resource_name,
		]
		command.extend(self._auth_args(context_id))
		ox_call(command, logger=self.logger)

	def run(self):
		if self.new and not self.new.get('mailPrimaryAddress', [''])[0]:
			self.logger.warn('No mailPrimaryAddress, exiting.')
			return

		if self.new and not self.old:
			self.create_resource(self.new_context_id)

		if self.new and self.old:
			if self.old_context_id == self.new_context_id:
				self.modify_resource(self.new_context_id)
			else:
				self.remove_resource(self.old_context_id)
				self.create_resource(self.new_context_id)

		if not self.new and self.old:
			self.remove_resource(self.old_context_id)


@ListenerQueue(name, stop_on_error=False)
def handler(dn, new, old, action=None):
	if listener_is_deactivated(name):
		return

	ox_resource_listener = OXResourceListener(name, old, new)
	ox_resource_listener.run()

if __name__ == '__main__':
	handler.process_pending()
